# -*- coding: utf-8 -*-
"""
Tests for the command line interface
"""

import os
import sys
import shutil
from subprocess import check_output
import tempfile
import pkg_resources
import copy
import fileinput
import re
import yaml
import unittest

from ephyviewer import mkQApp

import neurotic

import logging
logger = logging.getLogger(__name__)


class CLITestCase(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory(prefix='neurotic-')

        # save the current global config settings so they can be restored
        # during tear-down
        self.original_config = copy.deepcopy(neurotic.global_config)

        # reset global config settings to their factory defaults (use deepcopy
        # so that _global_config_factory_defaults is not changed during tests)
        neurotic.global_config.clear()
        neurotic.global_config.update(copy.deepcopy(
            neurotic._global_config_factory_defaults))

        # get parameters for the example
        self.example_file = pkg_resources.resource_filename(
            'neurotic', 'example/metadata.yml')
        self.example_dataset = 'Aplysia feeding'

        # make a copy of the example metadata file in the temp directory and
        # then add an additional dataset to it
        self.temp_file = shutil.copy(self.example_file, self.temp_dir.name)
        with open(self.temp_file) as f:
            metadata = yaml.safe_load(f)
        self.duplicate_dataset = 'zzz_alphabetically_last'
        metadata[self.duplicate_dataset] = copy.deepcopy(
            metadata[self.example_dataset])
        with open(self.temp_file, 'w') as f:
            f.write(yaml.dump(metadata))

    def tearDown(self):
        # remove the temp directory
        self.temp_dir.cleanup()

        # restore the original global config
        neurotic.global_config.clear()
        neurotic.global_config.update(self.original_config)

    def test_global_config_template(self):
        """Test that the global config template matches the factory defaults"""

        # load the template global config file by first copying it to the temp
        # directory, uncommenting every parameter, and then updating
        # global_config using the modified file
        self.template_global_config_file = pkg_resources.resource_filename(
            'neurotic', 'global_config_template.txt')
        self.temp_global_config_file = shutil.copy(
            self.template_global_config_file, self.temp_dir.name)
        with fileinput.input(self.temp_global_config_file, inplace=True) as f:
            for line in f:
                if re.match('#.*=.*', line):
                    line = line[1:]
                print(line, end='')  # stdout is redirected into the file
        neurotic.update_global_config_from_file(self.temp_global_config_file)

        self.assertEqual(neurotic.global_config,
                         neurotic._global_config_factory_defaults,
                         'global_config loaded from template global config '
                         'file differs from factory defaults')

    def test_cli_installed(self):
        """Test that the command line interface is installed"""
        self.assertIsNotNone(shutil.which('neurotic'), 'path to cli not found')

    def test_help(self):
        """Test that --help returns usage info"""
        argv = ['neurotic', '--help']
        out = check_output(argv)
        self.assertTrue(out.decode('utf-8').startswith('usage: neurotic'),
                        'help\'s stdout has unexpected content')

    def test_version(self):
        """Test that --version returns version info"""
        argv = ['neurotic', '--version']
        out = check_output(argv)
        self.assertTrue(out.decode('utf-8').startswith('neurotic'),
                        'version\'s stdout has unexpected content')

    def test_cli_defaults(self):
        """Test CLI default values match factory defaults"""
        argv = ['neurotic']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)

        # should match factory defaults because setUp() explicitly reset the
        # defaults to the factory defaults
        factory_defaults = neurotic._global_config_factory_defaults['defaults']
        self.assertEqual(win.do_toggle_debug_logging.isChecked(),
                         factory_defaults['debug'],
                         'debug setting has unexpected default')
        self.assertEqual(win.lazy, factory_defaults['lazy'],
                         'lazy setting has unexpected default')
        self.assertEqual(win.support_increased_line_width,
                         factory_defaults['thick_traces'],
                         'thick traces setting has unexpected default')
        self.assertEqual(win.show_datetime, factory_defaults['show_datetime'],
                         'show_datetime has unexpected default')
        self.assertEqual(win.ui_scale, factory_defaults['ui_scale'],
                         'ui_scale has unexpected default')
        self.assertEqual(win.theme, factory_defaults['theme'],
                         'theme has unexpected default')
        self.assertEqual(win.metadata_selector.file, self.example_file,
                         'file has unexpected default')
        self.assertEqual(win.metadata_selector._selection,
                         self.example_dataset,
                         'dataset has unexpected default')

    def test_debug(self):
        """Test that --debug enables logging of debug messages"""
        argv = ['neurotic', '--debug']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertTrue(win.do_toggle_debug_logging.isChecked(),
                        'debug logging disabled with --debug')

    def test_no_debug(self):
        """Test that --no-debug disables logging of debug messages"""
        argv = ['neurotic', '--no-debug']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertFalse(win.do_toggle_debug_logging.isChecked(),
                         'debug logging enabled with --no-debug')

    def test_lazy(self):
        """Test that --lazy enables lazy loading"""
        argv = ['neurotic', '--lazy']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertTrue(win.lazy, 'lazy loading disbled with --lazy')

    def test_no_lazy(self):
        """Test that --no-lazy disables lazy loading"""
        argv = ['neurotic', '--no-lazy']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertFalse(win.lazy, 'lazy loading enabled with --no-lazy')

    def test_thick_traces(self):
        """Test that --thick-traces enables support for thick traces"""
        argv = ['neurotic', '--thick-traces']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertTrue(win.support_increased_line_width,
                        'thick traces disabled with --thick-traces')

    def test_no_thick_traces(self):
        """Test that --no-thick-traces disables support for thick traces"""
        argv = ['neurotic', '--no-thick-traces']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertFalse(win.support_increased_line_width,
                         'thick traces enabled with --no-thick-traces')

    def test_show_datetime(self):
        """Test that --show-datetime enables display of real-world datetime"""
        argv = ['neurotic', '--show-datetime']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertTrue(win.show_datetime,
                        'datetime not displayed with --show-datetime')

    def test_no_show_datetime(self):
        """Test that --no-show-datetime hides the real-world datetime"""
        argv = ['neurotic', '--no-show-datetime']
        args = neurotic.parse_args(argv)
        app = mkQApp()
        win = neurotic.win_from_args(args)
        self.assertFalse(win.show_datetime,
                         'datetime displayed with --no-show-datetime')

    def test_ui_scale(self):
        """Test that --ui-scale changes the ui_scale"""
        app = mkQApp()

        for size in neurotic.available_ui_scales:
            argv = ['neurotic', '--ui-scale', size]
            args = neurotic.parse_args(argv)
            win = neurotic.win_from_args(args)
            self.assertEqual(win.ui_scale, size, 'unexpected scale')

    def test_theme(self):
        """Test that --theme changes the theme"""
        app = mkQApp()

        for theme in neurotic.available_themes:
            argv = ['neurotic', '--theme', theme]
            args = neurotic.parse_args(argv)
            win = neurotic.win_from_args(args)
            self.assertEqual(win.theme, theme, 'unexpected theme')

    def test_use_factory_defaults(self):
        """Test that --use-factory-defaults resets all defaults"""
        for k in neurotic.global_config['defaults']:
            neurotic.global_config['defaults'][k] = 'bad value'
        argv = ['neurotic', '--use-factory-defaults']
        args = neurotic.parse_args(argv)
        for k, v in neurotic._global_config_factory_defaults['defaults'].items():
            self.assertEqual(getattr(args, k), v,
                             f'args.{k} was not reset to factory default')

    def test_file(self):
        """Test that metadata file can be set"""
        argv = ['neurotic', self.temp_file]
        args = neurotic.parse_args(argv)
        win = neurotic.win_from_args(args)
        self.assertEqual(win.metadata_selector.file, self.temp_file,
                         'file was not changed correctly')

    def test_dataset(self):
        """Test that dataset can be set"""
        argv = ['neurotic', self.temp_file, self.duplicate_dataset]
        args = neurotic.parse_args(argv)
        win = neurotic.win_from_args(args)
        self.assertEqual(win.metadata_selector._selection,
                         self.duplicate_dataset,
                         'dataset was not changed correctly')

    def test_example_file_and_first_dataset_overrides(self):
        """Test that 'example' and 'first' open first dataset in example file"""
        neurotic.global_config['defaults']['file'] = 'some other file'
        neurotic.global_config['defaults']['dataset'] = 'some other dataset'
        argv = ['neurotic', 'example', 'first']
        args = neurotic.parse_args(argv)
        win = neurotic.win_from_args(args)
        self.assertEqual(win.metadata_selector.file, self.example_file,
                         'file was not changed correctly')
        self.assertEqual(win.metadata_selector._selection,
                         self.example_dataset,
                         'dataset was not changed correctly')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
The :mod:`neurotic.datasets.gdrive` module implements a class for downloading
files from Google Drive using paths, rather than file IDs or shareable links.

.. autoclass:: GoogleDriveDownloader
   :members:
"""

import os
import shutil
import urllib
from functools import reduce
from tqdm.auto import tqdm
from pydrive2.auth import GoogleAuth, LoadAuth
from pydrive2.drive import GoogleDrive

import logging
logger = logging.getLogger(__name__)


class GoogleDriveDownloader(GoogleDrive):
    """
    A class for downloading files from Google Drive using paths.

    Files can be specified for download using URL-like paths of the form

        gdrive://<drive name>/<folder 1>/<...>/<folder N>/<file name>

    The "<drive name>" may be "My Drive" for files located in a personal
    Google Drive, or it may be the name of a Shared Drive that the user has
    permission to access.

    Note that these URL-like paths are not equivalent to ordinary URLs
    associated with Google Drive files, such as shareable links, which are
    composed of pseudorandom file IDs and do not reveal anything about the name
    of the file or the folders containing it.

    This class can only download files that are uniquely identifiable by their
    paths. Google Drive does not require file or folder names to be unique, so
    two or more files or folders with identical names may coexist in a folder.
    Such files and folders cannot be distinguished by their paths, so they
    cannot be downloaded using this class. A download will fail while
    traversing the file tree if at any step there is more than one folder or
    file that matches the path.

    This class manages access authorization, optionally saving authorization
    tokens to a file so that the authorization flow does not need to be
    repeated in the future.

    The ``client_secret_file`` should be the path to a client secret file in
    JSON format, obtained from the `Google API Console
    <https://console.developers.google.com/>`_. The Drive API must be enabled
    for the corresponding client.

    If ``save_tokens=False``, the authorization flow (a request via web browser
    for permission to access Google Drive) will always run the first time a new
    instance of this class is used, and authorization will not persist after
    the instance is destroyed. If ``save_tokens=True`` and a file path is
    provided with ``tokens_file``, access/refresh tokens resulting from a
    successful authorization are stored in the file, and tokens are loaded from
    the file in the future, so that the authorization flow does not need to be
    repeated.
    """

    def __init__(self, client_secret_file, tokens_file=None, save_tokens=False):
        """
        Initialize a new GoogleDriveDownloader.
        """

        self.settings = {
            'client_config_file': client_secret_file,
            'oauth_scope': ['https://www.googleapis.com/auth/drive.readonly'],
            'save_credentials': save_tokens,
            'save_credentials_backend': 'file',
            'save_credentials_file': tokens_file
        }

        GoogleDrive.__init__(self, auth=self._create_auth())

    def _create_auth(self):
        """
        Create a GoogleAuth object with the correct settings.
        """
        auth = GoogleAuth()
        auth.settings.update(self.settings)
        return auth

    @LoadAuth
    def authorize(self):
        """
        Obtain tokens for reading the contents of a Google Drive account.

        If ``save_tokens=True``, tokens will be loaded from the ``tokens_file``
        if possible. If tokens cannot be restored this way, or if the loaded
        tokens have expired, an authorization flow will be initiated, prompting
        the user through a web browser to grant read-only privileges to the
        client associated with the ``client_secret_file``. When the
        authorization flow completes, if ``save_tokens=True``, the newly
        created tokens will be stored in the ``tokens_file`` for future use.

        Authorization is performed automatically when needed, but this method
        can be called directly to retrieve (and possibly store) tokens without
        initiating a download.
        """
        # the LoadAuth decorator does all the work
        return

    def deauthorize(self):
        """
        Forget tokens and delete the ``tokens_file``. The authorization flow
        will be required for the next download.
        """
        if os.path.exists(self.settings['save_credentials_file']):
            os.remove(self.settings['save_credentials_file'])
        del self.auth
        self.auth = self._create_auth()

    def is_authorized(self):
        """
        Get the current authorization state.
        """
        return (self.auth is not None and
                self.auth.credentials is not None and
                self.auth.service is not None)

    @LoadAuth
    def GetUserEmail(self):
        """
        Get the email address for the authorized Google Drive account.
        """
        return self.GetAbout()['user']['emailAddress']

    @LoadAuth
    def GetSharedDrivesList(self):
        """
        Return information about available Shared Drives.
        """
        # no PyDrive2 interface for this, so implement it here
        return self.auth.service.drives().list().execute(http=self.http)

    def download(self, gdrive_url, local_file, overwrite_existing=False, show_progress=True, bytes_per_chunk=1024*1024*5):
        """
        Download a file from Google Drive using a URL-like path beginning with
        "gdrive://".
        """
        if not overwrite_existing and os.path.exists(local_file):
            logger.info(f'Skipping {os.path.basename(local_file)} (already exists)')
            return

        logger.info(f'Downloading {os.path.basename(local_file)}')
        try:
            self._download_with_progress_bar(gdrive_url, local_file, show_progress=show_progress, bytes_per_chunk=bytes_per_chunk)
        except Exception as e:
            logger.error(f'Skipping {os.path.basename(local_file)} ({e})')
            raise

    def _download_with_progress_bar(self, gdrive_url, local_file, show_progress=True, bytes_per_chunk=1024*1024*5):
        """
        Download while showing a progress bar.
        """
        # TODO: bytes_per_chunk=1024*1024*100 (100 MiB) would match
        # googleapiclient.http.MediaIoBaseDownload's default chunk size and
        # seems to be significantly faster than smaller values, suggesting
        # chunk fetching incurs a large overhead. Unfortunately, such a large
        # chunk size would prevent the progress bar from updating frequently.
        # As a compromise, the chunk size used by this method is just 5 MiB,
        # which is a little larger than is ideal for progress reporting and yet
        # still noticeably slows downloads. Is there a better solution?

        # determine where to temporarily save the file during download
        temp_file = local_file + '.part'
        logger.debug(f'Temporarily downloading to {temp_file}')

        # create the containing directory if necessary
        if not os.path.exists(os.path.dirname(local_file)):
            os.makedirs(os.path.dirname(local_file))

        # locate the Google Drive file
        file_id = self._get_file_id(gdrive_url)
        if file_id is None:
            raise ValueError(f'error locating file on server for account "{self.GetUserEmail()}"')
        file = self.CreateFile({'id': file_id})

        try:
            with tqdm(total=int(file['fileSize']), unit='B', unit_scale=True) as pbar:
                def update_pbar(total_transferred, file_size):
                    pbar.n = total_transferred
                    pbar.update()
                file.GetContentFile(temp_file, callback=update_pbar, chunksize=bytes_per_chunk)

        except:
            # the download is likely incomplete, so delete the temporary file
            if os.path.exists(temp_file):
                os.remove(temp_file)

            # raise the exception so that it can be handled elsewhere
            raise

        else:
            # download completed, so move the temp file to the final location
            shutil.move(temp_file, local_file)

    def _get_file_id(self, gdrive_url):
        """
        Retrieve the Google Drive ID for the file specified by ``gdrive_url``.
        """
        # verify the url is of the right type
        scheme = urllib.parse.urlparse(gdrive_url).scheme
        if scheme != 'gdrive':
            raise ValueError(f'gdrive_url must begin with "gdrive://": {gdrive_url}')

        # extract drive name ("My Drive" or some Shared Drive) and file path
        drive_name = urllib.parse.urlparse(gdrive_url).netloc
        path = urllib.parse.urlparse(gdrive_url).path
        path = os.path.normpath(path).strip(os.sep).split(os.sep)

        # find the drive id from its name
        if not drive_name:
            raise ValueError('problem parsing drive name')
        elif drive_name == 'My Drive':
            drive_id = 'root'
        else:
            # search for all Shared Drives with a matching name
            drives = self.GetSharedDrivesList().get('items', [])
            drives = [drive for drive in drives if drive['name'] == drive_name]

            # make sure the drive is unique
            if len(drives) == 0:
                raise ValueError(f'drive "{drive_name}" not found on server for account "{self.GetUserEmail()}"')
            elif len(drives) > 1:
                raise ValueError(f'ambigous path, multiple drives with name "{drive_name}" exist on server for account "{self.GetUserEmail()}"')
            else:
                drive_id = drives[0]['id']

        # find the file id from its path by starting at the drive root and
        # recursively searching for the id of the next folder in the path
        file_id = reduce(self._get_child_id, path, drive_id)

        return file_id

    def _get_child_id(self, parent_id, child_name):
        """
        Retrieve the Google Drive ID for the file or folder named
        ``child_name`` located in a folder or drive with ID ``parent_id``.
        """
        # search for all files with a matching name and parent id
        items = self.ListFile({'q': f'title="{child_name}" and "{parent_id}" '
                                    'in parents and trashed=false'}).GetList()

        # make sure the file is unique
        if len(items) == 0:
            raise ValueError(f'file or folder "{child_name}" not found on server for account "{self.GetUserEmail()}"')
        elif len(items) > 1:
            raise ValueError(f'ambiguous path, multiple files or folders with the name "{child_name}" exist under their parent folder on server for account "{self.GetUserEmail()}"')
        else:
            child_id = items[0]['id']

        return child_id



# -*- coding: utf-8 -*-
"""
NEUROscience Tool for Interactive Characterization

Curate, visualize, annotate, and share your behavioral ephys data using Python
"""

import os
import sys
import shutil
import copy
import pkg_resources
import collections.abc
import logging
import logging.handlers
import toml

from .version import version as __version__
from .version import git_revision as __git_revision__


# set the user's directory for global settings file, logs, and more
neurotic_dir = os.path.join(os.path.expanduser('~'), '.neurotic')
if not os.path.exists(neurotic_dir):
    os.mkdir(neurotic_dir)


class FileLoggingFormatter(logging.Formatter):
    """
    A custom formatter for file logging
    """
    default_msec_format = '%s.%03d'  # use period radix point instead of comma in decimal seconds
    def format(self, record):
        if logging.getLogger(__name__).level <= logging.DEBUG:
            # include more detail if the logger level (not the record level) is
            # debug or lower
            self._style._fmt = '[%(asctime)s] [%(levelname)-8s] [%(threadName)-10s] [%(name)s:%(lineno)d (%(funcName)s)] %(message)s'
        else:
            self._style._fmt = '[%(asctime)s] [%(levelname)-8s] %(message)s'
        return super().format(record)

class StreamLoggingFormatter(logging.Formatter):
    """
    A custom formatter for stream logging
    """
    def format(self, record):
        if record.levelno == logging.INFO:
            # exclude the level name ("INFO") from common log records
            self._style._fmt = '[neurotic] %(message)s'
        else:
            self._style._fmt = '[neurotic] %(levelname)s: %(message)s'
        return super().format(record)


# set the file path for logging
log_file = os.path.join(neurotic_dir, 'neurotic-log.txt')

# set the default level for logging to INFO unless it was set to a custom level
# before importing the package
logger = logging.getLogger(__name__)
if logger.level == logging.NOTSET:
    default_log_level = logging.INFO
    logger.setLevel(default_log_level)
else:
    default_log_level = logger.level

# write log records to a file, rotating files if it exceeds 10 MB
logger_filehandler = logging.handlers.RotatingFileHandler(filename=log_file, maxBytes=10000000, backupCount=2)
logger_filehandler.setFormatter(FileLoggingFormatter())
logger.addHandler(logger_filehandler)
logger.info('===========================')        # file logger only
logger.info(f'Importing neurotic {__version__}')  # file logger only

# stream log records to stderr
logger_streamhandler = logging.StreamHandler(stream=sys.stderr)
logger_streamhandler.setFormatter(StreamLoggingFormatter())
logger.addHandler(logger_streamhandler)


global_config = {
    'defaults': {
        # defaults used by the command line interface
        'file': False,
        'dataset': False,
        'debug': False,
        'lazy': True,
        'thick_traces': False,
        'show_datetime': False,
        'ui_scale': 'medium',
        'theme': 'light',
    },
    'gdrive': {
        # parameters for Google Drive access
        'client_secret_file': os.path.join(neurotic_dir, 'gdrive-creds', 'client_secret.json'),
        'tokens_file': os.path.join(neurotic_dir, 'gdrive-creds', 'tokens.json'),
        'save_tokens': False,
    },
    'app': {
        'auto_check_for_updates': True,
    },
}

# keep a copy of the original config before it is modified
_global_config_factory_defaults = copy.deepcopy(global_config)

# the global config file is a text file in TOML format owned by the user that
# allows alternate defaults to be specified to replace those in global_config
global_config_file = os.path.join(neurotic_dir, 'neurotic-config.txt')

if not os.path.exists(global_config_file):
    # copy a template global config file containing commented-out defaults
    shutil.copy(
        pkg_resources.resource_filename(
            'neurotic', 'global_config_template.txt'),
        global_config_file)

def update_dict(d, d_new):
    """
    Recursively update the contents of a dictionary. Unlike dict.update(), this
    function preserves items in inner dictionaries that are absent from d_new.

    For example, if given

    >>> d = {'x': 0, 'inner': {'a': 1, 'b': 2}}
    >>> d_new = {'inner': {'c': 3}}

    then using d.update(d_new) will entirely replace d['inner'] with
    d_new['inner']:

    >>> d.update(d_new)
    >>> d == {'x': 0, 'inner': {'c': 3}}

    In contrast, update_dict(d, d_new) will preserve items found in d['inner']
    but not in d_new['inner']:

    >>> update_dict(d, d_new)
    >>> d == {'x': 0, 'inner': {'a': 1, 'b': 2, 'c': 3}}
    """
    for k_new, v_new in d_new.items():
        if isinstance(v_new, collections.abc.Mapping):
            d[k_new] = update_dict(d.get(k_new, {}), v_new)
        else:
            d[k_new] = v_new
    return d

def update_global_config_from_file(file=global_config_file):
    """
    Update the global_config dictionary with data from the global config file,
    using recursion to traverse nested dictionaries.
    """
    with open(file, 'r') as f:
        update_dict(global_config, toml.loads(f.read()))

try:
    update_global_config_from_file()
except Exception as e:
    logger.error(f'Ignoring global config file due to parsing error ({global_config_file}): {e}')


# create directories for storing Google Drive credentials if necessary
for file in [global_config['gdrive']['client_secret_file'],
             global_config['gdrive']['tokens_file']]:
    if file and not os.path.exists(os.path.dirname(file)):
        os.mkdir(os.path.dirname(file))


from .datasets import *
from .gui import *
from .scripts import *



# -*- coding: utf-8 -*-
"""
The :mod:`neurotic.scripts` module handles starting the app from the command
line. It also provides a convenience function for quickly launching the app
using minimal metadata or an existing Neo :class:`Block <neo.core.Block>`, and
another for starting a Jupyter server with an example notebook.

.. autofunction:: quick_launch

.. autofunction:: launch_example_notebook
"""

import os
import sys
import argparse
import subprocess
import pkg_resources

from ephyviewer import QT, mkQApp

from . import __version__, global_config, _global_config_factory_defaults, global_config_file, default_log_level
from .datasets.data import load_dataset
from .gui.config import EphyviewerConfigurator, available_themes, available_ui_scales
from .gui.standalone import MainWindow

import logging
logger = logging.getLogger(__name__)


def parse_args(argv):
    """

    """

    description = """
    neurotic lets you curate, visualize, annotate, and share your behavioral
    ephys data.
    """

    epilog = f"""
    Defaults for arguments and options can be changed in a global config file,
    {os.path.relpath(global_config_file, os.path.expanduser('~'))}, located in
    your home directory.
    """

    parser = argparse.ArgumentParser(description=description, epilog=epilog)

    defaults = global_config['defaults']
    parser.set_defaults(**defaults)

    parser.add_argument('file', nargs='?',
                        help='the path to a metadata YAML file'
                             f' (default: {"an example file" if defaults["file"] in [False, "example"] else defaults["file"] + "; to force the example, use: example"})')

    parser.add_argument('dataset', nargs='?',
                        help='the name of a dataset in the metadata file to '
                             'select initially'
                             f' (default: {"the first entry in the metadata file" if defaults["dataset"] in [False, "first"] else defaults["dataset"] + "; to force the first, use: - first"})')

    parser.add_argument('-V', '--version',
                        action='version',
                        version='neurotic {}'.format(__version__))

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--debug', dest='debug',
                       action='store_true',
                       help='enable detailed log messages for debugging'
                            f'{" (default)" if defaults["debug"] else ""}')
    group.add_argument('--no-debug', dest='debug',
                       action='store_false',
                       help='disable detailed log messages for debugging'
                            f'{" (default)" if not defaults["debug"] else ""}')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--lazy', dest='lazy',
                       action='store_true',
                       help='enable fast loading'
                            f'{" (default)" if defaults["lazy"] else ""}')
    group.add_argument('--no-lazy', dest='lazy',
                       action='store_false',
                       help='disable fast loading'
                            f'{" (default)" if not defaults["lazy"] else ""}')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--thick-traces', dest='thick_traces',
                       action='store_true',
                       help='enable support for traces with thick lines, '
                            'which has a performance cost'
                            f'{" (default)" if defaults["thick_traces"] else ""}')
    group.add_argument('--no-thick-traces', dest='thick_traces',
                       action='store_false',
                       help='disable support for traces with thick lines'
                            f'{" (default)" if not defaults["thick_traces"] else ""}')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--show-datetime', dest='show_datetime',
                       action='store_true',
                       help='display the real-world date and time, which '
                            'may be inaccurate depending on file type and '
                            'acquisition software'
                            f'{" (default)" if defaults["show_datetime"] else ""}')
    group.add_argument('--no-show-datetime', dest='show_datetime',
                       action='store_false',
                       help='do not display the real-world date and time'
                            f'{" (default)" if not defaults["show_datetime"] else ""}')

    parser.add_argument('--ui-scale', dest='ui_scale',
                        choices=available_ui_scales,
                        help='the scale of user interface elements, such as '
                             'text'
                             f' (default: {defaults["ui_scale"]})')

    parser.add_argument('--theme', dest='theme',
                        choices=available_themes,
                        help='a color theme for the GUI'
                             f' (default: {defaults["theme"]})')

    parser.add_argument('--use-factory-defaults',
                        action='store_true',
                        help='start with "factory default" settings, ignoring '
                             'other args and your global config file')

    group = parser.add_argument_group('alternative modes')
    group.add_argument('--launch-example-notebook',
                       action='store_true',
                       help='launch Jupyter with an example notebook '
                            'instead of starting the standalone app (other '
                            'args will be ignored)')

    args = parser.parse_args(argv[1:])

    if args.use_factory_defaults:
        # replace every argument with the factory default
        for k, v in _global_config_factory_defaults['defaults'].items():
            setattr(args, k, v)

    # this special value for the first positional argument can be used to skip
    # specifying any file and instead use the default file
    if args.file == '-':
        args.file = defaults['file']

    # these special values for the positional arguments can be used to override
    # the defaults set in the global config file with the defaults normally set
    # in the absence of global config file settings
    if args.file == 'example':
        args.file = _global_config_factory_defaults['defaults']['file']
    if args.dataset == 'first':
        args.dataset = _global_config_factory_defaults['defaults']['dataset']

    if args.debug:
        logger.parent.setLevel(logging.DEBUG)

        # lower the threshold for PyAV messages printed to the console from
        # critical to warning
        logging.getLogger('libav').setLevel(logging.WARNING)

        if not args.launch_example_notebook:
            # show only if Jupyter won't be launched, since the setting will
            # not carry over into the kernel started by Jupyter
            logger.debug('Debug messages enabled')

    else:
        # this should only be necessary if parse_args is called with --no-debug
        # after having previously enabled debug messages in the current session
        # (such as during unit testing)

        logger.parent.setLevel(default_log_level)

        # raise the threshold for PyAV messages printed to the console from
        # warning to critical
        logging.getLogger('libav').setLevel(logging.CRITICAL)

    logger.debug(f'Global config: {global_config}')
    logger.debug(f'Parsed arguments: {args}')

    return args

def win_from_args(args):
    """

    """

    win = MainWindow(file=args.file, initial_selection=args.dataset,
                     lazy=args.lazy, theme=args.theme, ui_scale=args.ui_scale,
                     support_increased_line_width=args.thick_traces,
                     show_datetime=args.show_datetime)
    return win

def launch_example_notebook():
    """
    Start a Jupyter server and open the example notebook.
    """

    path = pkg_resources.resource_filename('neurotic',
                                           'example/example-notebook.ipynb')
    out = None

    # check whether Jupyter is installed
    try:
        out = subprocess.Popen(['jupyter', 'notebook', '--version'],
                               stdout=subprocess.PIPE).communicate()[0]
    except FileNotFoundError as e:
        logger.error('Unable to verify Jupyter is installed using "jupyter '
                     'notebook --version". Is it installed?')

    if out:
        # run Jupyter on the example notebook
        try:
            out = subprocess.Popen(['jupyter', 'notebook', path],
                                   stdout=subprocess.PIPE).communicate()[0]
        except FileNotFoundError as e:
            logger.error(f'Unable to locate the example notebook at {path}')

def main():
    """

    """

    args = parse_args(sys.argv)
    if args.launch_example_notebook:
        launch_example_notebook()
    else:
        logger.info('Loading user interface')
        app = mkQApp()
        splash = QT.QSplashScreen(QT.QPixmap(':/neurotic-logo-300.png'))
        splash.show()
        win = win_from_args(args)
        win.show()
        splash.finish(win)
        logger.info('Ready')
        app.exec_()

def quick_launch(metadata={}, blk=None, lazy=True):
    """
    Load data, configure the GUI, and launch the app with one convenient
    function.

    This function allows *neurotic* to be used easily in interactive sessions
    and scripts. For example, dictionaries can be passed as metadata:

    >>> metadata = {'data_file': 'data.axgx'}
    >>> neurotic.quick_launch(metadata=metadata)

    An existing Neo :class:`Block <neo.core.Block>` can be passed directly:

    >>> neurotic.quick_launch(blk=my_neo_block)

    This function is equivalent to the following:

    >>> blk = load_dataset(metadata, blk, lazy=lazy)
    >>> ephyviewer_config = EphyviewerConfigurator(metadata, blk, lazy=lazy)
    >>> ephyviewer_config.show_all()
    >>> ephyviewer_config.launch_ephyviewer()
    """

    # make sure this matches the docstring after making changes
    blk = load_dataset(metadata, blk, lazy=lazy)
    ephyviewer_config = EphyviewerConfigurator(metadata, blk, lazy=lazy)
    ephyviewer_config.show_all()
    ephyviewer_config.launch_ephyviewer()
